import os
import sys
import zipfile

import gdown

def download_data(path):
    file_id = "1efHsY16pxK0lBD2gYCgCTnv1Swstq771"
    url = f"https://drive.google.com/uc?id={file_id}"
    data_zip = os.path.join(path, "data.zip")
    gdown.download(url, data_zip, quiet=False)
    
    with zipfile.ZipFile(data_zip, "r") as zip_ref:
        zip_ref.extractall(path)
    return
    
def main():
    download_data("./")
    return
    
if __name__ == "__main__":
    main()    
    sys.exit()
    


import re
import os
import sys
import unicodedata

import numpy as np
import tensorflow as tf

from ckipneutools import model_ws
from ckipneutools import model_pos
from ckipneutools import model_ner

os.environ["CUDA_VISIBLE_DEVICES"] = ""
_whitespace_pattern = re.compile("[\s]+")

def construct_dictionary(word_to_weight):
    length_word_weight = {}
    
    for word, weight in word_to_weight.items():
        if not word: continue
        try:
            weight = float(weight)
        except ValueError:
            continue
        length = len(word)
        if length not in length_word_weight:
            length_word_weight[length] = {}
        length_word_weight[length][word] = weight
        
    length_word_weight = sorted(length_word_weight.items())
    
    return length_word_weight
    
class WS:
    def __init__(self, data_dir):
        config = model_ws.Config()
        config.name = "model_asbc_Att-0_BiLSTM-cross-2-500_batch128-run1"
        config.attention_heads = 0
        config.is_cross_bilstm = True
        config.layers = 2
        config.hidden_d = 500
        config.w_token_to_vector, config.w_embedding_d = _load_embedding(os.path.join(data_dir, "embedding_character"))
        
        with tf.Graph().as_default():
            model = model_ws.Model(config)
            model.sess = tf.Session()
            model.sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver()
            saver.restore(model.sess, os.path.join(data_dir, "model_ws", config.name))
            
        self.model = model
        return
        
    def __del__(self):
        self.model.sess.close()
        return
        
    def __call__(
            self,
            sentence_list,
            
            recommend_dictionary = {},
            coerce_dictionary = {},
            
            sentence_segmentation = False,
            segment_delimiter_set = {",", "。", ":", "?", "!", ";"},
            
            character_normalization = True,
            batch_sentences = 2048,
            batch_characters = 16384,
        ):
        
        # Character normalization
        if character_normalization:
            raw_sentence_list = sentence_list
            sentence_list = []
            normal2raw_list = []
            for raw_sentence in raw_sentence_list:
                sentence, normal_to_raw_index = _normalize_sentence(raw_sentence)
                sentence_list.append(sentence)
                normal2raw_list.append(normal_to_raw_index)
            segment_delimiter_set = {unicodedata.normalize("NFKD", word) for word in segment_delimiter_set}
            
        # Sentence segmentation
        segment_list = []
        segment_to_sentence_index = []
        for sentence_index, sentence in enumerate(sentence_list):
            if sentence_segmentation:
                partial_segment_list = _segment_sentence(sentence, segment_delimiter_set)
            else:
                partial_segment_list = [sentence]
            for segment in partial_segment_list:
                segment_list.append(segment)
                segment_to_sentence_index.append(sentence_index)
                
        # Sequence-label prediction
        batch_list = _get_ws_batch_list(segment_list, batch_sentences, batch_characters)
        seq_segment_list = [None] * len(segment_list)
        for batch in batch_list:
            index_list, sample_list = zip(*batch)
            if len(sample_list[0][0]) == 0:
                parital_seq_segment_list = [[] for sample in sample_list]
            else:
                parital_seq_segment_list = self.model.predict_label_for_a_batch(sample_list)
            for partial_index, index in enumerate(index_list):
                seq_segment_list[index] = parital_seq_segment_list[partial_index]
                
        # Undo sentence segmentation
        seq_sentence_list = []
        seq_sentence = []
        sentence_index = 0
        for segment_index, seq_segment in enumerate(seq_segment_list):
            assert len(segment_list[segment_index]) == len(seq_segment)
            if sentence_index != segment_to_sentence_index[segment_index]:
                seq_sentence_list.append(seq_sentence)
                seq_sentence = []
                sentence_index = segment_to_sentence_index[segment_index]
            seq_sentence += seq_segment
        seq_sentence_list.append(seq_sentence)
        
        # Undo character normalization
        if character_normalization:
            sentence_list = raw_sentence_list
            raw_seq_sentence_list = []
            for seq_sentence, normal_to_raw_index in zip(seq_sentence_list, normal2raw_list):
                raw_seq_sentence = []
                for index, label in enumerate(seq_sentence):
                    if normal_to_raw_index[index] < 0: continue
                    raw_seq_sentence.append(label)
                raw_seq_sentence_list.append(raw_seq_sentence)
            seq_sentence_list = raw_seq_sentence_list
            
        # Word segmentation
        word_sentence_list = []
        for sentence, seq_segment in zip(sentence_list, seq_sentence_list):
            word_sentence = _get_word_sentence_from_seq_sentence(sentence, seq_segment)
            word_sentence = _run_word_segmentation_with_dictionary(word_sentence, recommend_dictionary, coerce_dictionary)
            word_sentence_list.append(word_sentence)
            
        return word_sentence_list
        
class POS:
    def __init__(self, data_dir):
        config = model_pos.Config()
        config.name = "model_asbc_Att-0_BiLSTM-2-500_batch256-run1"
        config.attention_heads = 0
        config.layers = 2
        config.hidden_d = 500
        config.c_token_to_vector, config.c_embedding_d = _load_embedding(os.path.join(data_dir, "embedding_character"))
        config.w_token_to_vector, config.w_embedding_d = _load_embedding(os.path.join(data_dir, "embedding_word"))
        config.label_list, config.label_to_index = _read_pos_list(os.path.join(data_dir, "model_pos", "label_list.txt"))
        config.output_d = len(config.label_list)
        
        with tf.Graph().as_default():
            model = model_pos.Model(config)
            model.sess = tf.Session()
            model.sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver()
            saver.restore(model.sess, os.path.join(data_dir, "model_pos", config.name))
        
        self.model = model
        return
        
    def __del__(self):
        self.model.sess.close()
        return
        
    def __call__(
            self,
            sentence_list,
            
            sentence_segmentation = True,
            segment_delimiter_set = {",", "。", ":", "?", "!", ";"},
            
            character_normalization = True,
            batch_sentences = 2048,
            batch_characters = 16384,
        ):
        
        # Character normalization
        if character_normalization:
            raw_sentence_list = sentence_list
            sentence_list = []
            for raw_sentence in raw_sentence_list:
                sentence = []
                for raw_word in raw_sentence:
                    word = unicodedata.normalize("NFKD", raw_word)
                    sentence.append(word)
                sentence_list.append(sentence)
                
        # Sentence segmentation
        segment_list = []
        segment_to_sentence_index = []
        for sentence_index, sentence in enumerate(sentence_list):
            if sentence_segmentation:
                partial_segment_list = _segment_word_sentence(sentence, segment_delimiter_set)
            else:
                partial_segment_list = [sentence]
            for segment in partial_segment_list:
                segment_list.append(segment)
                segment_to_sentence_index.append(sentence_index)
                
        # POS prediction
        batch_list = _get_pos_batch_list(segment_list, batch_sentences, batch_characters)
        pos_segment_list = [None] * len(segment_list)
        for batch in batch_list:
            index_list, sample_list = zip(*batch)
            if len(sample_list[0][0]) == 0:
                parital_pos_segment_list = [[] for sample in sample_list]
            else:
                parital_pos_segment_list = self.model.predict_label_for_a_batch(sample_list)
            for partial_index, index in enumerate(index_list):
                pos_segment_list[index] = parital_pos_segment_list[partial_index]
                
        # Undo sentence segmentation
        pos_sentence_list = []
        pos_sentence = []
        sentence_index = 0
        for segment_index, pos_segment in enumerate(pos_segment_list):
            assert len(segment_list[segment_index]) == len(pos_segment)
            _force_whitespace_tagging(segment_list[segment_index], pos_segment)
            
            if sentence_index != segment_to_sentence_index[segment_index]:
                pos_sentence_list.append(pos_sentence)
                pos_sentence = []
                sentence_index = segment_to_sentence_index[segment_index]
            pos_sentence += pos_segment
        pos_sentence_list.append(pos_sentence)
        
        return pos_sentence_list
        
class NER:
    def __init__(self, data_dir):
        config = model_ner.Config()
        config.name = "model_ontochinese_Att-0_BiLSTM-2-500_batch128-run1"
        config.attention_heads = 0
        config.layers = 2
        config.hidden_d = 500
        config.c_token_to_vector, config.c_embedding_d = _load_embedding(os.path.join(data_dir, "embedding_character"))
        config.w_token_to_vector, config.w_embedding_d = _load_embedding(os.path.join(data_dir, "embedding_word"))
        _, config.pos_to_index = _read_pos_list(os.path.join(data_dir, "model_ner", "pos_list.txt"))
        config.w_feature_d = len(config.pos_to_index)
        config.label_list, config.label_to_index = _read_entity_type_list(os.path.join(data_dir, "model_ner", "label_list.txt"))
        config.output_d = len(config.label_list)
        
        with tf.Graph().as_default():
            model = model_ner.Model(config)
            model.sess = tf.Session()
            model.sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver()
            saver.restore(model.sess, os.path.join(data_dir, "model_ner", config.name))
            
        self.model = model
        return
        
    def __del__(self):
        self.model.sess.close()
        return
        
    def __call__(
            self,
            word_sentence_list,
            pos_sentence_list,
            
            character_normalization = True,
            batch_sentences = 2048,
            batch_characters = 16384,
        ):
        
        # Character normalization
        if character_normalization:
            raw_word_sentence_list = word_sentence_list
            word_sentence_list = []
            normal2raw_list = []
            for raw_word_sentence in raw_word_sentence_list:
                word_sentence = [unicodedata.normalize("NFKD", raw_word) for raw_word in raw_word_sentence]
                sentence, normal_to_raw_index = _normalize_sentence("".join(raw_word_sentence))
                assert sentence == "".join(word_sentence)
                word_sentence_list.append(word_sentence)
                normal2raw_list.append(normal_to_raw_index)
                
        # Sequence-label prediction
        batch_list = _get_ner_batch_list(word_sentence_list, pos_sentence_list, batch_sentences, batch_characters)
        label_sentence_list = [None] * len(word_sentence_list)
        for batch in batch_list:
            index_list, sample_list = zip(*batch)
            if len(sample_list[0][0]) == 0:
                parital_label_sentence_list = [[] for sample in sample_list]
            else:
                parital_label_sentence_list = self.model.predict_label_for_a_batch(sample_list)
            for partial_index, index in enumerate(index_list):
                label_sentence_list[index] = parital_label_sentence_list[partial_index]
                
        # Undo character normalization
        if character_normalization:
            word_sentence_list = raw_word_sentence_list
            raw_label_sentence_list = []
            for label_sentence, normal_to_raw_index in zip(label_sentence_list, normal2raw_list):
                raw_label_sentence = []
                for index, label in enumerate(label_sentence):
                    if normal_to_raw_index[index] < 0:
                        if label[-1]!="E" or raw_label_sentence[-1]=="O": continue
                        entity_type, seq_type = raw_label_sentence[-1].split(":")
                        if entity_type != label[:-2]: continue
                        if seq_type == "B":
                            raw_label_sentence[-1] = entity_type + ":S"
                        elif seq_type == "I":
                            raw_label_sentence[-1] = entity_type + ":E"
                        continue
                    raw_label_sentence.append(label)
                raw_label_sentence_list.append(raw_label_sentence)
            label_sentence_list = raw_label_sentence_list
        
        # Entity mention collection
        entity_sentence_list = []
        for word_sentence, label_sentence in zip(word_sentence_list, label_sentence_list):
            entity_sentence = _get_entity_set(word_sentence, label_sentence)
            entity_sentence_list.append(entity_sentence)
            
        return entity_sentence_list
        
def _load_embedding(embedding_dir):
    token_file = os.path.join(embedding_dir, "token_list.npy")
    token_list = np.load(token_file)
    
    vector_file = os.path.join(embedding_dir, "vector_list.npy")
    vector_list = np.load(vector_file)
    
    token_to_vector = dict(zip(token_list, vector_list))
    d = vector_list.shape[1]
    return token_to_vector, d
    
def _read_pos_list(label_list_file):
    with open(label_list_file, "r") as f:
        line_list = f.read().splitlines()
        
    label_list = []
    for line in line_list:
        label, count = line.split(" ")
        label_list.append(label)
    label_to_index = {label: index for index, label in enumerate(label_list)}
    
    return label_list, label_to_index
    
def _read_entity_type_list(label_list_file):
    with open(label_list_file, "r") as f:
        line_list = f.read().splitlines()
        
    label_list = ["O"]
    for line in line_list:
        label, count = line.split(" ")
        label_list.append(label+":S")
        label_list.append(label+":B")
        label_list.append(label+":I")
        label_list.append(label+":E")
    label_to_index = {label: index for index, label in enumerate(label_list)}
    
    return label_list, label_to_index
    
def _normalize_sentence(raw_sentence):
    normal_sentence = unicodedata.normalize("NFKD", raw_sentence)
    
    normal_to_raw_index = []
    for raw_index, raw_string in enumerate(raw_sentence):
        normal_string = unicodedata.normalize("NFKD", raw_string)
        normal_index = len(normal_to_raw_index)
        assert normal_string == normal_sentence[normal_index:normal_index+len(normal_string)]
        normal_to_raw_index.append(raw_index)
        for _ in normal_string[1:]:
            normal_to_raw_index.append(-1)
            
    return normal_sentence, normal_to_raw_index
    
def _segment_sentence(sentence, delimiter_set):
    if not sentence:
        return [""]
        
    segment_list = []
    segment = ""
    for c in sentence:
        segment += c
        if c in delimiter_set:
            segment_list.append(segment)
            segment = ""
    if segment:
        segment_list.append(segment)
    return segment_list
    
def _segment_word_sentence(sentence, delimiter_set):
    if not sentence:
        return [[]]
        
    segment_list = []
    segment = []
    for word in sentence:
        segment.append(word)
        if word in delimiter_set:
            segment_list.append(segment)
            segment = []
    if segment:
        segment_list.append(segment)
    return segment_list
    
def _get_ws_batch_list(sentence_list, batch_sentences, batch_characters):
    index_sentence_list = sorted(
        enumerate(sentence_list),
        key = lambda index_sentence: len(index_sentence[1])
    )
    
    empty_sentence_batch = []
    for index, sentence in index_sentence_list:
        if len(sentence) > 0: break
        empty_sentence_batch.append((index, (sentence,[])))
    empty_sentences = len(empty_sentence_batch)
    if empty_sentence_batch:    
        batch_list = [empty_sentence_batch]
    else:
        batch_list = []
        
    batch = []
    for index, sentence in index_sentence_list[empty_sentences:]:
        batch.append((index, (sentence,["B" for c in sentence])))
        if len(batch)>=batch_sentences or len(batch)*len(batch[-1][1][0])>=batch_characters:
            batch_list.append(batch)
            batch = []
    if batch:
        batch_list.append(batch)
        
    return batch_list
    
def _get_pos_batch_list(sentence_list, batch_sentences, batch_characters):
    index_sentence_list = sorted(
        enumerate(sentence_list),
        key = lambda index_sentence: sum(len(word) for word in index_sentence[1])
    )
    
    empty_sentence_batch = []
    for index, sentence in index_sentence_list:
        if len(sentence) > 0: break
        empty_sentence_batch.append((index, (sentence,[])))
    empty_sentences = len(empty_sentence_batch)
    if empty_sentence_batch:    
        batch_list = [empty_sentence_batch]
    else:
        batch_list = []
        
    batch = []
    for index, sentence in index_sentence_list[empty_sentences:]:
        batch.append((index, (sentence,["D" for word in sentence])))
        if len(batch)>=batch_sentences or len(batch)*sum(len(word) for word in batch[-1][1][0])>=batch_characters:
            batch_list.append(batch)
            batch = []
    if batch:
        batch_list.append(batch)
        
    return batch_list
    
def _get_ner_batch_list(word_sentence_list, pos_sentence_list, batch_sentences, batch_characters):
    label_sentence_list = [
        [
            "O"
            for word in word_sentence
                for character in word
        ]
        for word_sentence in word_sentence_list
    ]
    
    index_sample_list = sorted(
        enumerate(zip(word_sentence_list, pos_sentence_list, label_sentence_list)),
        key = lambda i_wpl: sum(len(word) for word in i_wpl[1][0])
    )
    
    empty_sentence_batch = []
    for index, sample in index_sample_list:
        if len(sample[2]) > 0: break
        empty_sentence_batch.append((index, sample))
    empty_sentences = len(empty_sentence_batch)
    if empty_sentence_batch:
        batch_list = [empty_sentence_batch]
    else:
        batch_list = []
        
    batch = []
    for index, sample in index_sample_list[empty_sentences:]:
        batch.append((index, sample))
        if len(batch)>=batch_sentences or len(batch)*sum(len(word) for word in batch[-1][1][0])>=batch_characters:
            batch_list.append(batch)
            batch = []
    if batch:
        batch_list.append(batch)
        
    return batch_list
    
def _get_word_sentence_from_seq_sentence(sentence, seq_sentence):
    assert len(sentence) == len(seq_sentence)
    if not sentence:
        return []
        
    word_sentence = []
    word = sentence[0]
    for character, label in zip(sentence[1:], seq_sentence[1:]):
        if label == "B":
            word_sentence.append(word)
            word = ""
        word += character
    word_sentence.append(word)
    
    return word_sentence
    
def _get_forced_chunk_set(sentence, length_word_weight):
    
    chunk_to_weight = {}
    
    for i in range(len(sentence)):
        for length, word_to_weight in length_word_weight:
            word = sentence[i:i+length]
            if word in word_to_weight:
                chunk_to_weight[(i, i+length)] = word_to_weight[word]
                
    chunk_set = set()
    empty_sentence = [True] * len(sentence)
    
    for (l, r), w in sorted(chunk_to_weight.items(), key=lambda x: (x[1], x[0][1]-x[0][0]), reverse=True):
        empty = True
        for i in range(l, r):
            if not empty_sentence[i]:
                empty = False
                break
        if not empty: continue
        
        chunk_set.add((l,r))
        for i in range(l, r):
            empty_sentence[i] = False
            
    return chunk_set
    
def _soft_force_seq_sentence(forced_chunk_set, seq_sentence):
    for l, r in sorted(forced_chunk_set):
        if seq_sentence[l] != "B": continue
        if r<len(seq_sentence) and seq_sentence[r] != "B": continue
        for i in range(l+1,r):
            seq_sentence[i] = "I"
    return
    
def _hard_force_seq_sentence(forced_chunk_set, seq_sentence):
    for l, r in sorted(forced_chunk_set):
        seq_sentence[l] = "B"
        for i in range(l+1,r):
            seq_sentence[i] = "I"
        if r < len(seq_sentence):
            seq_sentence[r] = "B"
    return
    
def _run_word_segmentation_with_dictionary(word_sentence, recommend_dictionary, coerce_dictionary):
    sentence = "".join(word_sentence)
    seq_sentence = []
    for word in word_sentence:
        seq_sentence.append("B")
        for character in word[1:]:
            seq_sentence.append("I")
            
    recommend_chunk_set = _get_forced_chunk_set(sentence, recommend_dictionary)
    _soft_force_seq_sentence(recommend_chunk_set, seq_sentence)
    coerce_chunk_set = _get_forced_chunk_set(sentence, coerce_dictionary)
    _hard_force_seq_sentence(coerce_chunk_set, seq_sentence)
    
    word_sentence = _get_word_sentence_from_seq_sentence(sentence, seq_sentence)
    return word_sentence
    
def _force_whitespace_tagging(word_sentence, pos_sentence):
    for i, word in enumerate(word_sentence):
        if _whitespace_pattern.fullmatch(word) is not None:
            pos_sentence[i] = "WHITESPACE"
    return
    
def _get_entity_set(word_sentence, label_sentence):
    sentence = "".join(word_sentence)
    assert len(sentence) == len(label_sentence)
    
    entity_set = set()
    l = None
    for i, label in enumerate(label_sentence):
        if label == "O":
            l = None
        else:
            entity_type, seq_type = label.split(":")
            if seq_type == "S":
                entity_set.add(
                    (i, i+1, entity_type, sentence[i:i+1])
                )
                l = None
            elif seq_type == "B":
                l = i
            elif seq_type == "E":
                if l is not None:
                    entity_set.add(
                        (l, i+1, entity_type, sentence[l:i+1])
                    )
                    l = None
    return entity_set
    
def main():
    return
    
if __name__ == "__main__":
    main()
    sys.exit()
    


import sys
import string

import numpy as np
import tensorflow as tf

class Config:
    def __init__(self):
        self.name = "cross-bilstm"
        
        # Character
        self.c_embedding_d = 300
        self.c_feature_d = 6
        
        # Word
        self.w_embedding_d = 300
        self.w_feature_d = 0
        
        # BiLSTM
        self.hidden_d = 500
        self.layers = 2
        
        # Attention
        self.attention_heads = 0
        
        # Classifier
        self.output_d = 2
        self.label_list = ["NP", "VP"]
        self.label_to_index = {"NP": 0, "VP": 1}
        
        # Optimization
        self.max_gradient_norm = 5
        self.learning_rate = 1e-3
        self.keep_rate = 0.65
        return
        
class Cross_BiLSTM:
    def __init__(self, name="cross-bilstm", input_d=300, hidden_d=100, layers=2):
        self.name = name
        self.kr = tf.placeholder(tf.float32, [])
        
        self.f_cell_list = []
        self.b_cell_list = []
        for layer in range(layers):
            is_top_cell = (layer==layers-1)
            current_input_d = input_d if layer==0 else hidden_d*2
            self.f_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell, f"lstm_cell_{layer+1}")
            )
            self.b_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell, f"lstm_cell_{layer+1}")
            )
        return
        
    def create_cell(self, input_d, hidden_d, is_top_cell, name):
        cell = tf.nn.rnn_cell.LSTMCell(hidden_d, name=name)
        cell = tf.nn.rnn_cell.DropoutWrapper(
            cell,
            input_keep_prob = self.kr,
            state_keep_prob = self.kr,
            output_keep_prob = self.kr if is_top_cell else 1.0,
            variational_recurrent = True,
            input_size = input_d,
            dtype = tf.float32,
        )
        return cell
        
    def transform(self, X, L):
        """BiLSTM transform function.
        
        X: [batch, length, input_d]
        L: [batch]
        H: [batch, length, hidden_d*2]
        S_fw: [layers, 2, batch, hidden_d]
        S_bw: [layers, 2, batch, hidden_d]
        """
        H = X
        S_fw = []
        S_bw = []
        with tf.variable_scope(self.name):
            for f_cell, b_cell in zip(self.f_cell_list, self.b_cell_list):
                # top_output: [2, batch, length, hidden_d]
                #    axis0: forward/backward
                # last_state: [2, 2, batch, hidden_d]
                #    axis0: forward/backward
                #    axis1: LSTM c/h
                top_output, last_state = tf.nn.bidirectional_dynamic_rnn(
                    f_cell,
                    b_cell,
                    H,
                    sequence_length = L,
                    dtype = tf.float32,
                )
                H = tf.concat(top_output, 2) # [batch, length, hidden_d*2]
                S_fw.append(last_state[0])
                S_bw.append(last_state[1])
        return H, S_fw, S_bw
        
class Att:
    def __init__(self, name="att", heads=5, dk_head=60, dv_head=60):
        self.heads = heads
        self.QP_list = []
        self.KP_list = []
        self.VP_list = []
        for i in range(heads):
            self.QP_list.append(
                tf.layers.Dense(dk_head, use_bias=False, name=f"{name}/QP_{i}")
            )
            self.KP_list.append(
                tf.layers.Dense(dk_head, use_bias=False, name=f"{name}/KP_{i}")
            )
            self.VP_list.append(
                tf.layers.Dense(dv_head, use_bias=False, name=f"{name}/VP_{i}")
            )
        return
        
    def attention_transform(self, Q, K, V, L=None):
        """
        Q: [batch, q, dk]
        K: [batch, m, dk]
        V: [batch, m, dv]
        L: [batch]
        C: [batch, q, dv]
        """
        # Attention score
        dk = tf.cast(tf.shape(K)[2], tf.float32)
        score = tf.matmul(Q, K, transpose_b=True) # [batch, q, m]
        score = score/tf.sqrt(dk)                 # [batch, q, m]
        
        # Mask score
        if L is not None:
            mask = tf.sequence_mask(L)                                     # [batch, m]
            mask = tf.expand_dims(mask, 1)                                 # [batch, 1, m]
            mask = tf.tile(mask, [1, tf.shape(score)[1], 1])               # [batch, q, m]
            worst_score = tf.ones_like(score) * tf.constant(float("-inf")) # [batch, q, m]
            score = tf.where(mask, score, worst_score)                     # [batch, q, m]
            
        # Context vector
        alpha = tf.nn.softmax(score, 2) # [batch, q, m]
        C = tf.matmul(alpha, V)         # [batch, q, dv]
        return C
        
    def transform(self, Q, K, V, L=None):
        """
        Q: [batch, q, dq]
        K: [batch, m, dk]
        V: [batch, m, dv]
        C: [batch, q, heads * dv_head]
        """
        batch = tf.shape(Q)[0]
        q = tf.shape(Q)[1]
        m = tf.shape(K)[1]
        
        C_list = []
        for i in range(self.heads):
            Q_head = self.QP_list[i](Q) # [batch, q, dk_head]
            K_head = self.KP_list[i](K) # [batch, m, dk_head]
            V_head = self.VP_list[i](V) # [batch, m, dv_head]
            
            C_head = self.attention_transform(Q_head, K_head, V_head, L=L) # [batch, q, dv_head]
            C_list.append(C_head)
            
        C = tf.concat(C_list, 2) # [batch, q, heads * dv_head]
        return C
        
class Model:
    def __init__(self, config):
        """Contruct Tensowflow graph."""
        
        self.create_hyper_parameter(config)
        self.create_embedding()
        self.create_encoder()
        self.create_classifier()
        return
        
    def create_hyper_parameter(self, config):
        """Add attributes of config to self."""
        
        for parameter in dir(config):
            if parameter[0] == "_": continue
            setattr(self, parameter, getattr(config, parameter))
        return
        
    def create_embedding(self):
        """Create trainable unknown vectors."""
        with tf.variable_scope(self.name, initializer=tf.random_normal_initializer(stddev=0.1)):
            self.unknown_c_v = tf.get_variable("unknown_c_v", [1, self.c_embedding_d+self.c_feature_d])
            self.unknown_w_v = tf.get_variable("unknown_w_v", [1, self.w_embedding_d+self.w_feature_d])
        return
        
    def create_encoder(self):
        """Create an Att-BiLSTM encoder."""
        
        # Input sequence                     batch, s_l_max
        self.s_l = tf.placeholder(  tf.int32, [None])
        self.c_k = tf.placeholder(  tf.int32, [None, None])
        self.c_v = tf.placeholder(tf.float32, [None, None, self.c_embedding_d+self.c_feature_d])
        self.w_k = tf.placeholder(  tf.int32, [None, None])
        self.w_v = tf.placeholder(tf.float32, [None, None, self.w_embedding_d+self.w_feature_d])
        
        batch  = tf.shape(self.c_v)[0]
        s_l_max = tf.shape(self.c_v)[1]
        
        # Character
        c_known     = tf.equal(tf.reshape(self.c_k, [batch*s_l_max]), 1)                         # [batch*s_l_max]
        c_known_v   = tf.reshape(self.c_v, [batch*s_l_max, self.c_embedding_d+self.c_feature_d]) # [batch*s_l_max, c_embedding_d+c_feature_d]
        c_unknown_v = tf.tile(self.unknown_c_v, [batch*s_l_max, 1])                              # [batch*s_l_max, c_embedding_d+c_feature_d]
        c_v         = tf.where(c_known, c_known_v, c_unknown_v)                                  # [batch*s_l_max, c_embedding_d+c_feature_d]
        c_v         = tf.reshape(c_v, [batch, s_l_max, self.c_embedding_d+self.c_feature_d])     # [batch, s_l_max, c_embedding_d+c_feature_d]
        
        # Word
        w_known     = tf.equal(tf.reshape(self.w_k, [batch*s_l_max]), 1)                         # [batch*s_l_max]
        w_known_v   = tf.reshape(self.w_v, [batch*s_l_max, self.w_embedding_d+self.w_feature_d]) # [batch*s_l_max, w_embedding_d+w_feature_d]
        w_unknown_v = tf.tile(self.unknown_w_v, [batch*s_l_max, 1])                              # [batch*s_l_max, w_embedding_d+w_feature_d]
        w_v         = tf.where(w_known, w_known_v, w_unknown_v)                                  # [batch*s_l_max, w_embedding_d+w_feature_d]
        w_v         = tf.reshape(w_v, [batch, s_l_max, self.w_embedding_d+self.w_feature_d])     # [batch, s_l_max, w_embedding_d+w_feature_d]
        
        cw_v         = tf.concat([c_v, w_v], axis=2) # [batch, s_l_max, c_embedding_d+c_feature_d+w_embedding_d+w_feature_d]
        
        # BiLSTM
        with tf.variable_scope(self.name):
            self.bilstm = Cross_BiLSTM(
                input_d = self.c_embedding_d + self.c_feature_d + self.w_embedding_d + self.w_feature_d,
                hidden_d = self.hidden_d,
                layers = self.layers,
            )
            cw_h, _, _ = self.bilstm.transform(cw_v, self.s_l) # [batch, s_l_max, hidden_d*2]
            
        # Multi-head attention
        if self.attention_heads == 0:
            self.cw_a = cw_h
        else:
            with tf.variable_scope(self.name):
                head_d = int(self.w_hidden_d * 2 / self.attention_heads)
                att = Att(
                    heads = self.attention_heads,
                    dk_head = head_d,
                    dv_head = head_d,
                )
                cw_c = att.transform(cw_h, cw_h, cw_h, L=self.s_l) # [batch, s_l_max, hidden_d*2]
            self.cw_a = tf.concat([cw_h, cw_c], 2) # [batch, s_l_max, hidden_d*4]
        return
        
    def create_classifier(self):
        """Create a layer to compute predictions and loss."""
        
        # Output
        # self.logits are not actually logits, because a tf.nn.log_softmax() is omitted
        output_layer = tf.layers.Dense(
            self.output_d,
            use_bias = True,
            name = "output_layer",
        )
        with tf.variable_scope(self.name):
            self.logits = output_layer(self.cw_a) # [batch, s_l_max, output_d]
            
        # Loss
        self.o_i    = tf.placeholder(  tf.int32, [None, None]) # [batch, s_l_max]
        self.o_mask = tf.placeholder(tf.float32, [None, None]) # [batch, s_l_max]
        
        cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels = self.o_i,
            logits = self.logits,
        )
        self.loss = tf.reduce_sum(cross_entropy * self.o_mask) / tf.reduce_sum(self.o_mask)
        
        # Optimization
        optimizer = tf.contrib.opt.NadamOptimizer(learning_rate=self.learning_rate)
        gv_list = optimizer.compute_gradients(self.loss)
        g_list, v_list = zip(*gv_list)
        g_list = [
            tf.convert_to_tensor(g) if isinstance(g, tf.IndexedSlices) else g
            for g in g_list
        ]
        if self.max_gradient_norm > 0:
            g_list, _ = tf.clip_by_global_norm(g_list, self.max_gradient_norm)
        self.update_op = optimizer.apply_gradients(zip(g_list, v_list))
        return
        
    def get_word_feature(self, word):
        feature_list = [0] * self.w_feature_d
        return feature_list
        
    def get_character_feature(self, character, ci):
        assert len(character) == 1
        feature_list = [0] * self.c_feature_d
        
        if character in string.ascii_uppercase:
            feature_list[0] = 1
            
        elif character in string.ascii_lowercase:
            feature_list[1] = 1
            
        elif character in string.digits:
            feature_list[2] = 1
            
        elif character in string.punctuation:
            feature_list[3] = 1
            
        elif ord("\u4e00") <= ord(character) <= ord("\u9fff"):
            feature_list[4] = 1
            
        if ci > 0:
            feature_list[5] = 1
            
        return feature_list
        
    def get_formatted_input(self, sample_list):
        """Preprocessing: extract Numpy arrays for TensorFlow input placeholders.
        
        sample_list: [sample]
            sample: (word_sequence, label_sequence)
        """
        batch = len(sample_list)
        s_l_max = max(
            sum(
                len(word)
                for word in word_sequence
            )
            for word_sequence, _ in sample_list
        )
        
        s_l    = np.zeros([batch                                              ], dtype=np.int32)
        c_k    = np.zeros([batch, s_l_max,                                    ], dtype=np.int32)
        c_v    = np.zeros([batch, s_l_max, self.c_embedding_d+self.c_feature_d], dtype=np.float32)
        w_k    = np.zeros([batch, s_l_max                                     ], dtype=np.int32)
        w_v    = np.zeros([batch, s_l_max, self.w_embedding_d+self.w_feature_d], dtype=np.float32)
        o_i    = np.zeros([batch, s_l_max                                     ], dtype=np.int32)
        o_mask = np.zeros([batch, s_l_max                                     ], dtype=np.float32)
        
        for b, (word_sequence, label_sequence) in enumerate(sample_list):
            s_l[b] = sum(len(word) for word in word_sequence)
            t = 0
            
            assert len(word_sequence) == len(label_sequence)
            for word, label in zip(word_sequence, label_sequence):
                if word in self.w_token_to_vector:
                    word_known = True
                    word_vector = np.concatenate((
                        self.w_token_to_vector[word],
                        self.get_word_feature(word),
                    ))
                else:
                    word_known = False
                o_i[b,t] = self.label_to_index[label]
                o_mask[b,t] = 1
                
                for ci, character in enumerate(word):
                    if character in self.c_token_to_vector:
                        c_k[b,t] = 1
                        c_v[b,t] = np.concatenate((
                            self.c_token_to_vector[character],
                            self.get_character_feature(character, ci),
                        ))
                    if word_known:
                        w_k[b,t] = 1
                        w_v[b,t] = word_vector
                    t += 1
                    
        return (
            s_l,
            c_k, c_v,
            w_k, w_v,
            o_i, o_mask,
        )
        
    def compute_loss_for_a_batch(self, sample_list, update_model=False):
        """Compute loss for a mini-batch, update model parameters if required.
        
        Return loss
            loss: the sum of cross-entropy across all tokens.
        """
        (
            s_l,
            c_k, c_v,
            w_k, w_v,
            o_i, o_mask,
        ) = self.get_formatted_input(sample_list)
        
        if update_model:
            output_list = [self.loss, self.update_op]
        else:
            output_list = [self.loss]
        
        result_list = self.sess.run(
            output_list,
            feed_dict = {
                self.bilstm.kr: self.keep_rate if update_model else 1.0,
                self.s_l: s_l,
                self.c_k: c_k,
                self.c_v: c_v,
                self.w_k: w_k,
                self.w_v: w_v,
                self.o_i: o_i,
                self.o_mask: o_mask,
            }
        )
        return result_list[0]
        
    def predict_label_for_a_batch(self, sample_list):
        """Generate token label sequence for each sample in the mini-batch.
        
        prediction_list: [prediction]
            prediction: [token_label]
        """
        (
            s_l,
            c_k, c_v,
            w_k, w_v,
            _, o_mask,
        ) = self.get_formatted_input(sample_list)
        
        logits = self.sess.run( # [batch, s_l_max, output_d]
            self.logits,
            feed_dict = {
                self.bilstm.kr: 1.0,
                self.s_l: s_l,
                self.s_l: s_l,
                self.c_k: c_k,
                self.c_v: c_v,
                self.w_k: w_k,
                self.w_v: w_v,
            }
        )
        
        labelindex_sequence_list = np.argmax(logits, axis=2) # [batch, s_l_max]
        prediction_list = []
        
        for b, labelindex_sequence in enumerate(labelindex_sequence_list):
            prediction = []
            for t, label_index in enumerate(labelindex_sequence[:s_l[b]]):
                if o_mask[b,t] == 1:
                    label = self.label_list[label_index]
                    prediction.append(label)
            prediction_list.append(prediction)
            
        return prediction_list
        
def main():
    config = Config()
    model = Model(config)
    tf_config = tf.ConfigProto()
    tf_config.gpu_options.per_process_gpu_memory_fraction = 0.45
    with tf.Session(config=tf_config) as sess:
        sess.run(tf.global_variables_initializer())
        for v in tf.trainable_variables():
            print(v)
    return
    
if __name__ == "__main__":
    main()
    sys.exit()
    


import sys
import string

import numpy as np
import tensorflow as tf

class Config:
    def __init__(self):
        self.name = "cross-bilstm"
        
        # Word
        self.w_embedding_d = 300
        self.w_feature_d = 4
        
        # BiLSTM
        self.is_cross_bilstm = True
        self.hidden_d = 100
        self.layers = 2
        
        # Attention
        self.attention_heads = 0
        
        # Classifier
        self.output_d = 2
        self.tokenlabel_list = ["B", "I"]
        self.tokenlabel_to_index = {"B": 0, "I": 1}
        
        # Optimization
        self.max_gradient_norm = 5
        self.learning_rate = 1e-3
        self.keep_rate = 0.65
        return
        
class BiLSTM:
    def __init__(self, name="bilstm", input_d=300, hidden_d=100, layers=2):
        self.name = name
        self.kr = tf.placeholder(tf.float32, [])
        
        f_cell_list = []
        b_cell_list = []
        for layer in range(layers):
            is_top_cell = (layer==layers-1)
            current_input_d = input_d if layer==0 else hidden_d
            f_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell)
            )
            b_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell)
            )
        self.f_cell = tf.nn.rnn_cell.MultiRNNCell(f_cell_list)
        self.b_cell = tf.nn.rnn_cell.MultiRNNCell(b_cell_list)
        return
        
    def create_cell(self, input_d, hidden_d, is_top_cell):
        cell = tf.nn.rnn_cell.LSTMCell(hidden_d)
        cell = tf.nn.rnn_cell.DropoutWrapper(
            cell,
            input_keep_prob = self.kr,
            state_keep_prob = self.kr,
            output_keep_prob = self.kr if is_top_cell else 1.0,
            variational_recurrent = True,
            input_size = input_d,
            dtype = tf.float32,
        )
        return cell
        
    def transform(self, X, L):
        """BiLSTM transform function.
        
        X: [batch, length, input_d]
        L: [batch]
        H: [batch, length, hidden_d*2]
        """
        with tf.variable_scope(self.name):
            # top_output: [2, batch, length, hidden_d], axis0: forward/backward
            # last_state: [2, layers, 2, batch, hidden_d], axis0: forward/backward, axis2: LSTM c/h
            top_output, last_state = tf.nn.bidirectional_dynamic_rnn(
                self.f_cell,
                self.b_cell,
                X,
                sequence_length = L,
                dtype = tf.float32,
            )
        H = tf.concat(top_output, 2) # [batch, length, hidden_d*2]
        return H
        
class Cross_BiLSTM:
    def __init__(self, name="cross-bilstm", input_d=300, hidden_d=100, layers=2):
        self.name = name
        self.kr = tf.placeholder(tf.float32, [])
        
        self.f_cell_list = []
        self.b_cell_list = []
        for layer in range(layers):
            is_top_cell = (layer==layers-1)
            current_input_d = input_d if layer==0 else hidden_d*2
            self.f_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell, f"lstm_cell_{layer+1}")
            )
            self.b_cell_list.append(
                self.create_cell(current_input_d, hidden_d, is_top_cell, f"lstm_cell_{layer+1}")
            )
        return
        
    def create_cell(self, input_d, hidden_d, is_top_cell, name):
        cell = tf.nn.rnn_cell.LSTMCell(hidden_d, name=name)
        cell = tf.nn.rnn_cell.DropoutWrapper(
            cell,
            input_keep_prob = self.kr,
            state_keep_prob = self.kr,
            output_keep_prob = self.kr if is_top_cell else 1.0,
            variational_recurrent = True,
            input_size = input_d,
            dtype = tf.float32,
        )
        return cell
        
    def transform(self, X, L):
        """BiLSTM transform function.
        
        X: [batch, length, input_d]
        L: [batch]
        H: [batch, length, hidden_d*2]
        """
        H = X
        with tf.variable_scope(self.name):
            for f_cell, b_cell in zip(self.f_cell_list, self.b_cell_list):
                # top_output: [2, batch, length, hidden_d], axis0: forward/backward
                # last_state: [2, layers, 2, batch, hidden_d], axis0: forward/backward, axis2: LSTM c/h
                top_output, last_state = tf.nn.bidirectional_dynamic_rnn(
                    f_cell,
                    b_cell,
                    H,
                    sequence_length = L,
                    dtype = tf.float32,
                )
                H = tf.concat(top_output, 2) # [batch, length, hidden_d*2]
        return H
        
class Att:
    def __init__(self, name="att", heads=5, dk_head=60, dv_head=60):
        self.heads = heads
        self.QP_list = []
        self.KP_list = []
        self.VP_list = []
        for i in range(heads):
            self.QP_list.append(
                tf.layers.Dense(dk_head, use_bias=False, name=f"{name}/QP_{i}")
            )
            self.KP_list.append(
                tf.layers.Dense(dk_head, use_bias=False, name=f"{name}/KP_{i}")
            )
            self.VP_list.append(
                tf.layers.Dense(dv_head, use_bias=False, name=f"{name}/VP_{i}")
            )
        return
        
    def attention_transform(self, Q, K, V, L=None):
        """
        Q: [batch, q, dk]
        K: [batch, m, dk]
        V: [batch, m, dv]
        L: [batch]
        C: [batch, q, dv]
        """
        # Attention score
        dk = tf.cast(tf.shape(K)[2], tf.float32)
        score = tf.matmul(Q, K, transpose_b=True) # [batch, q, m]
        score = score/tf.sqrt(dk)                 # [batch, q, m]
        
        # Mask score
        if L is not None:
            mask = tf.sequence_mask(L)                                     # [batch, m]
            mask = tf.expand_dims(mask, 1)                                 # [batch, 1, m]
            mask = tf.tile(mask, [1, tf.shape(score)[1], 1])               # [batch, q, m]
            worst_score = tf.ones_like(score) * tf.constant(float("-inf")) # [batch, q, m]
            score = tf.where(mask, score, worst_score)                     # [batch, q, m]
            
        # Context vector
        alpha = tf.nn.softmax(score, 2) # [batch, q, m]
        C = tf.matmul(alpha, V)         # [batch, q, dv]
        return C
        
    def transform(self, Q, K, V, L=None):
        """
        Q: [batch, q, dq]
        K: [batch, m, dk]
        V: [batch, m, dv]
        C: [batch, q, heads * dv_head]
        """
        batch = tf.shape(Q)[0]
        q = tf.shape(Q)[1]
        m = tf.shape(K)[1]
        
        C_list = []
        for i in range(self.heads):
            Q_head = self.QP_list[i](Q) # [batch, q, dk_head]
            K_head = self.KP_list[i](K) # [batch, m, dk_head]
            V_head = self.VP_list[i](V) # [batch, m, dv_head]
            
            C_head = self.attention_transform(Q_head, K_head, V_head, L=L) # [batch, q, dv_head]
            C_list.append(C_head)
            
        C = tf.concat(C_list, 2) # [batch, q, heads * dv_head]
        return C
        
class Model:
    def __init__(self, config):
        """Contruct Tensowflow graph."""
        
        self.create_hyper_parameter(config)
        self.create_embedding()
        self.create_encoder()
        self.create_classifier()
        return
        
    def create_hyper_parameter(self, config):
        """Add attributes of config to self."""
        
        for parameter in dir(config):
            if parameter[0] == "_": continue
            setattr(self, parameter, getattr(config, parameter))
        return
        
    def create_embedding(self):
        """Create a trainable unknown vector."""
        with tf.variable_scope(self.name, initializer=tf.random_normal_initializer(stddev=0.1)):
            self.unknown_w_v = tf.get_variable("unknown_w_v", [1, self.w_embedding_d+self.w_feature_d])
        return
        
    def create_encoder(self):
        """Create an Att-BiLSTM encoder."""
        
        # Input sequence                              [batch, length]
        self.input_length = tf.placeholder(  tf.int32, [None])
        self.w_k          = tf.placeholder(  tf.int32, [None, None])
        self.w_v          = tf.placeholder(tf.float32, [None, None, self.w_embedding_d+self.w_feature_d])
        
        batch  = tf.shape(self.w_v)[0]
        length = tf.shape(self.w_v)[1]
        
        # Word
        w_known     = tf.equal(tf.reshape(self.w_k, [batch*length]), 1)                         # [batch * length]
        w_known_v   = tf.reshape(self.w_v, [batch*length, self.w_embedding_d+self.w_feature_d]) # [batch * length, w_embedding_d+w_feature_d]
        w_unknown_v = tf.tile(self.unknown_w_v, [batch*length, 1])                              # [batch * length, w_embedding_d+w_feature_d]
        w_v         = tf.where(w_known, w_known_v, w_unknown_v)                                 # [batch * length, w_embedding_d+w_feature_d]
        w_v         = tf.reshape(w_v, [batch, length, self.w_embedding_d+self.w_feature_d])     # [batch,  length, w_embedding_d+w_feature_d]
        
        # BiLSTM
        with tf.variable_scope(self.name):
            if self.is_cross_bilstm:
                self.bilstm = Cross_BiLSTM(
                    input_d = self.w_embedding_d + self.w_feature_d,
                    hidden_d = self.hidden_d,
                    layers = self.layers,
                )
            else:
                self.bilstm = BiLSTM(
                    input_d = self.w_embedding_d + self.w_feature_d,
                    hidden_d = self.hidden_d,
                    layers = self.layers,
                )
            w_h = self.bilstm.transform(w_v, self.input_length) # [batch, length, hidden_d*2]
            
        # Multi-head attention
        if self.attention_heads == 0:
            self.w_a = w_h
        else:
            with tf.variable_scope(self.name):
                head_d = int(self.hidden_d * 2 / self.attention_heads)
                att = Att(
                    heads = self.attention_heads,
                    dk_head = head_d,
                    dv_head = head_d,
                )
                w_c = att.transform(w_h, w_h, w_h, L=self.input_length) # [batch, length, hidden_d*2]
            self.w_a = tf.concat([w_h, w_c], 2) # [batch, length, hidden_d*4]
        return
        
    def create_classifier(self):
        """Create a layer to compute predictions and loss."""
        
        # Output
        # self.logits are not actually logits, because a tf.nn.log_softmax() is omitted
        output_layer = tf.layers.Dense(
            self.output_d,
            use_bias = True,
            name = "output_layer",
        )
        with tf.variable_scope(self.name):
            self.logits = output_layer(self.w_a) # [batch, length, output_d]
            
        # Loss
        self.o_i    = tf.placeholder(  tf.int32, [None, None]) # [batch, length]
        self.o_mask = tf.placeholder(tf.float32, [None, None]) # [batch, length]
        
        cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels = self.o_i,
            logits = self.logits,
        )
        self.loss = tf.reduce_sum(cross_entropy * self.o_mask) / tf.reduce_sum(self.o_mask)
        
        # Optimization
        optimizer = tf.contrib.opt.NadamOptimizer(learning_rate=self.learning_rate)
        gv_list = optimizer.compute_gradients(self.loss)
        g_list, v_list = zip(*gv_list)
        g_list = [
            tf.convert_to_tensor(g) if isinstance(g, tf.IndexedSlices) else g
            for g in g_list
        ]
        if self.max_gradient_norm > 0:
            g_list, _ = tf.clip_by_global_norm(g_list, self.max_gradient_norm)
        self.update_op = optimizer.apply_gradients(zip(g_list, v_list))
        return
        
    def get_word_feature(self, word):
        assert len(word) == 1
        feature_list = [0] * self.w_feature_d
        
        if word in string.ascii_uppercase:
            feature_list[0] = 1
            
        elif word in string.ascii_lowercase:
            feature_list[1] = 1
            
        elif word in string.digits:
            feature_list[2] = 1
            
        elif word in string.punctuation:
            feature_list[3] = 1
            
        return feature_list
        
    def get_formatted_input(self, sample_list):
        """Preprocessing: extract Numpy arrays for TensorFlow input placeholders.
        
        sample_list: [sample]
            sample: (word_sequence, tokenlabel_sequence)
        """
        batch = len(sample_list)
        length = max(len(word_sequence) for word_sequence, _ in sample_list)
        
        input_length = np.zeros([batch                                             ], dtype=np.int32)
        w_k          = np.zeros([batch, length                                     ], dtype=np.int32)
        w_v          = np.zeros([batch, length, self.w_embedding_d+self.w_feature_d], dtype=np.float32)
        o_i          = np.zeros([batch, length                                     ], dtype=np.int32)
        o_mask       = np.zeros([batch, length                                     ], dtype=np.float32)
        
        for b, (word_sequence, tokenlabel_sequence) in enumerate(sample_list):
            input_length[b] = len(word_sequence)
            
            for t, word in enumerate(word_sequence):
                if word in self.w_token_to_vector:
                    w_k[b,t] = 1
                    w_v[b,t] = np.concatenate((
                        self.w_token_to_vector[word],
                        self.get_word_feature(word),
                    ))
                    
                o_i[b,t] = self.tokenlabel_to_index[tokenlabel_sequence[t]]
                o_mask[b,t] = 1
                
        return (
            input_length,
            w_k, w_v,
            o_i, o_mask,
        )
        
    def compute_loss_for_a_batch(self, sample_list, update_model=False):
        """Compute loss for a mini-batch, update model parameters if required.
        
        Return loss
            loss: the sum of cross-entropy across all tokens.
        """
        (
            input_length,
            w_k, w_v,
            o_i, o_mask,
        ) = self.get_formatted_input(sample_list)
        
        if update_model:
            output_list = [self.loss, self.update_op]
        else:
            output_list = [self.loss]
        
        result_list = self.sess.run(
            output_list,
            feed_dict = {
                self.bilstm.kr: self.keep_rate if update_model else 1.0,
                self.input_length: input_length,
                self.w_k: w_k,
                self.w_v: w_v,
                self.o_i: o_i,
                self.o_mask: o_mask,
            }
        )
        return result_list[0]
        
    def predict_label_for_a_batch(self, sample_list):
        """Generate token label sequence for each sample in the mini-batch.
        
        prediction_list: [prediction]
            prediction: [token_label]
        """
        (
            input_length,
            w_k, w_v,
            _, _,
        ) = self.get_formatted_input(sample_list)
        
        logits = self.sess.run( # [batch, length, output_d]
            self.logits,
            feed_dict = {
                self.bilstm.kr: 1.0,
                self.input_length: input_length,
                self.w_k: w_k,
                self.w_v: w_v,
            }
        )
        
        labelindex_sequence_list = np.argmax(logits, axis=2) # [batch, length]
        prediction_list = []
        
        for b, labelindex_sequence in enumerate(labelindex_sequence_list):
            prediction = []
            for label_index in labelindex_sequence[:input_length[b]]:
                token_label = self.tokenlabel_list[label_index]
                prediction.append(token_label)
            prediction[0] = "B"
            prediction_list.append(prediction)
            
        return prediction_list
        
def main():
    config = Config()
    model = Model(config)
    tf_config = tf.ConfigProto()
    tf_config.gpu_options.per_process_gpu_memory_fraction = 0.45
    with tf.Session(config=tf_config) as sess:
        sess.run(tf.global_variables_initializer())
        for v in tf.trainable_variables():
            print(v)
    return
    
if __name__ == "__main__":
    main()
    sys.exit()
    